from struct import pack
from shutil import copyfile
import mmap
import array
from parse import parse
from crc_checksums import crc64ecma, crc32castagnoli

_mapping = None
_checksums = None
_duplicate_tables = None
_correct_checksums = None
_correct_duplicate_tables = None

def init_module(packed_data_len, mapping_path, checksum_path, duplicate_tables_path, correct_checksums, correct_duplicate_tables): # type: (str, int, str, str, str, bool, bool) -> None
    global _duplicate_tables
    global __mapping_path
    global _checksums
    global _correct_checksums
    global _correct_duplicate_tables
    global _mapping
    _mapping = ReFSMapping(mapping_path, packed_data_len)
    _checksums = ReFSChecksumChains(_mapping, checksum_path)
    _duplicate_tables = ReFSDuplicateTables(duplicate_tables_path)
    _correct_checksums = correct_checksums
    _correct_duplicate_tables = correct_duplicate_tables


class ReFSMapping(object):
    TYPE_SUPERBLOCK = 'SUPB'
    TYPE_CHECKPOINT = 'CHKP'
    TYPE_OTHER = 'STD'

    def __init__(self, mapping_file = None, size_of_packed = None): # type: (str, int) -> None
        self._mapping = list()
        self._supbs = list()
        self._chkps = list()

        if mapping_file:
            self.__init_with_file(mapping_file, size_of_packed)


    def __get_mapping_file_version(self, mapping_file):
        with open(mapping_file, "r") as mf:
            for line in mf:
                items = line.split("-")
                return 1 if len(items) == 3 else 2


    def __parse_mapping_v1(self, mapping_file, size_of_packed):
        packed_addr = None
        real_addr = None
        with open(mapping_file, "r") as mf:
            for line in mf:
                if packed_addr is not None:
                    new_packed_addr, _, _ = parse("{} - {} - {}", line)
                    new_packed_addr = int(new_packed_addr, 16)
                    size = new_packed_addr - packed_addr
                    self.add_translation_entry(packed_addr, real_addr, size, mtype)

                packed_addr, real_addr, mtype = parse("{} - {} - {}", line)
                packed_addr = int(packed_addr, 16)
                real_addr = int(real_addr, 16)

            self.add_translation_entry(packed_addr, real_addr, size_of_packed - packed_addr, mtype)


    def __parse_mapping_v2(self, mapping_file):
        with open(mapping_file, "r") as mf:
            for line in mf:
                real_addr, packed_addr, size, mtype = parse("{} - {} - {} - {}", line)
                packed_addr = int(packed_addr, 16)
                real_addr = int(real_addr, 16)
                size = int(size, 16)
                self.add_translation_entry(packed_addr, real_addr, size, mtype)


    def __init_with_file(self, mapping_file, size_of_packed):
        mapping_version = self.__get_mapping_file_version(mapping_file)
        if mapping_version == 1:
            assert size_of_packed != None, "We need to know the total size of packed payload when using mapping file format v1"
            self.__parse_mapping_v1(mapping_file, size_of_packed)
        else:
            self.__parse_mapping_v2(mapping_file)


    def add_translation_entry(self, packed_address, real_address, size, mtype): # type: (int, int, int, str) -> none
        assert size >= 0x1000
        entry_tuple = (packed_address, real_address, size)
        self._mapping.append(entry_tuple)
        if mtype == ReFSMapping.TYPE_SUPERBLOCK:
            self._supbs.append(entry_tuple)
        if mtype == ReFSMapping.TYPE_CHECKPOINT:
            self._chkps.append(entry_tuple)



    def get_superblocks(self): # type: () -> List[Tuple[int, int, int]]
        return list(self._supbs)


    def get_checkpoints(self): # type: () -> List[Tuple[int, int, int]]
        return list(self._chkps)


    def translate_to_real_address(self, packed_address, size): # type: (int, int) -> Tuple[Tuple[int, int], Optional[Tuple[int, int]]]
        assert size < 0x1000

        first_return = None
        second_return = None
        for packed_addr, real_addr, translation_size in self._mapping:
            # this saves the follower of the first entry in case the size is not big enough
            if first_return:
                second_return = real_addr
                break

            offset = packed_address - packed_addr
            if offset < 0 or offset >= translation_size:
                continue
            #first translation entry found
            if offset + size <= translation_size:
                return ((real_addr + offset, size), None)
            else:
                first_return = (real_addr + offset, translation_size - offset)

        # Here is the case that we need two translation because of the size requirement
        return (first_return, (second_return, size - first_return[1]))

    def translate_to_packed_address(self, real_address): # type: (int) -> int
        for packed_addr, real_addr, size in self._mapping:
            offset = real_address - real_addr
            if offset < 0:
                continue
            if offset >= size:
                continue
            return packed_addr + offset
        assert False, "we should not reach this point"


    def get_addresses_to_pack(self): # type: () -> List[Tuple[int, int]]
        returnvalue = list()
        for packed_addr, real_addr, translation_size in self._mapping:
            returnvalue.append((real_addr, translation_size))
        return returnvalue

    def get_addresses_to_expand(self): # type: () -> List[Tuple[int, int, int]]
        # returned tuples are packed_offset, expanded_offset, size
        return list(self._mapping)

    def get_page_start(self, packed_addr): # type: (int) -> int
        for packed_page_addr, _, size in self._mapping:
            offset = packed_addr - packed_page_addr
            if offset >= 0 and offset < size:
                return packed_page_addr
        assert False, "page start address should be found in this setup"

    def get_page_start_ra(self, real_addr): # type: (int) -> int
        for _, real_p_addr, size in self._mapping:
            offset = real_addr - real_p_addr
            if offset >= 0 and offset < size:
                return real_p_addr
        assert False, "page start address should be found in this setup"

    def get_page_size(self, packed_page_addr): # type: (int) -> int
        for packed_page_addr_l, _, size in self._mapping:
            if packed_page_addr_l == packed_page_addr:
                return size
        assert False, "page start address should be found in this setup"


class ReFSChecksumChains(object):
    TYPE_CRC_32 = 1
    TYPE_CRC_64 = 2

    class PageNode(object):
        def __init__(self, checksum_addr_packed, page_start_packed, ctype, clength, mapping = None): # type: (int, int, str, int, ReFSMapping) -> None
            self.checksumed_at = checksum_addr_packed
            self.page_start = page_start_packed
            self.length = clength

            self.ctype = -1
            if ctype == "CRC64":
                self.ctype = ReFSChecksumChains.TYPE_CRC_64
            if ctype == "CRC32":
                self.ctype = ReFSChecksumChains.TYPE_CRC_32

            if mapping:
                self.checksumed_at_real = mapping.translate_to_real_address(self.checksumed_at, 1)[0][0]
                self.page_start_real = mapping.translate_to_real_address(self.page_start, 1)[0][0]
                self.page_size = mapping.get_page_size(self.page_start)
            else:
                self.checksumed_at_real = None
                self.page_start_real = None
                self.page_size = None
            self.parent_pages = None

        def __str__(self): # type: () -> str
            if self.parent_pages:
                if len(self.parent_pages) == 1:
                    return "(%s %s) -> %s" % (hex(self.page_start), hex(self.checksumed_at) ,str(self.parent_pages[0]))
                else:
                    return "(%s %s) -> %s" % (hex(self.page_start), hex(self.checksumed_at) ,str(self.parent_pages))
            else:
                return "(%s %s)" % (hex(self.page_start), hex(self.checksumed_at))

        def __repr__(self): # type: () -> str
            return self.__str__()

        def set_self_ref(self, self_ref_addr, self_ref_size, mapping = None): # type: (int, int, ReFSMapping) -> None
            self.self_ref_addr = self_ref_addr
            self.self_ref_size = self_ref_size
            if mapping:
                self.self_ref_addr_real = mapping.translate_to_real_address(self_ref_addr, 1)[0][0]
            else:
                self.self_ref_addr_real = None



    def __init__(self, mapping, checksum_file = None): # type: (ReFSMapping, str) -> None
        self.mapping = mapping
        self.SUPB_self_ref = list()
        self.CHKP_self_ref = list()
        if checksum_file:
            self.__init_checksums(checksum_file)

    def __init_checksums(self, checksum_file): # type: (str) -> None
        references = list()
        with open(checksum_file, "r") as chksum_file:
            for line in chksum_file:
                if len(line.split("-")) == 6:
                    c_addr, target_addr, ctype, clength, self_ref_addr, self_ref_len = parse("{} - {} - {} - {} - {} - {}", line)
                    c_addr = int(c_addr, 16)
                    target_addr = int(target_addr, 16)
                    clength = int(clength)
                    self_ref_addr = int(self_ref_addr, 16)
                    self_ref_len = int(self_ref_len)
                    self.add_self_desc(c_addr, target_addr, ctype, clength, self_ref_addr, self_ref_len)
                else:
                    ref_addr, target_addr, ctype, clength = parse("{} - {} - {} - {}", line)
                    ref_addr = int(ref_addr, 16)
                    target_addr = int(target_addr, 16)
                    clength = int(clength)
                    references.append((ref_addr, target_addr, ctype, clength))

        self.build_chains(references)


    def add_self_desc(self, checksum_addr, target_addr, ctype, clength, self_ref_addr, self_ref_length): # type: (int, int, str, int, int, int) -> None
        node = ReFSChecksumChains.PageNode(checksum_addr, target_addr, ctype, clength, self.mapping)
        node.set_self_ref(self_ref_addr, self_ref_length, self.mapping)
        node.parent_pages = []
        for packed, real, _ in self.mapping.get_superblocks():
            if target_addr == packed:
                self.SUPB_self_ref.append(node)
                return

        self.CHKP_self_ref.append(node)


    def __recursive_build(self, checksums, checksum_addr, mapping): # type: (List[Tuple[int, int, str, int]], int, ReFSMapping) -> List[ReFSChecksumChains.PageNode]
        page_addr_of_chksm = mapping.get_page_start(checksum_addr)
        checksum_nodes = list()

        # check first if the checksum is in checkpoint. thats the irregular case
        for checkpoint_page in self.CHKP_self_ref:
            if checkpoint_page.page_start == page_addr_of_chksm:
                checksum_nodes.append(checkpoint_page)
                return checksum_nodes

        # this loop will find checksums for the page
        for checksum_addr_l, target_page, ctype, clength in checksums:
            if target_page != page_addr_of_chksm:
                continue
            # we found a checksum for the page
            node = ReFSChecksumChains.PageNode(checksum_addr_l, target_page, ctype, clength, mapping)
            checksum_nodes.append(node)
            parent_nodes = self.__recursive_build(checksums, checksum_addr_l, mapping)
            node.parent_pages = parent_nodes

        return checksum_nodes


    def build_chains(self, packed_pages): # type: (List[Tuple[int, int, str, int]]) -> None
        self._chains = dict()
        self._chains_pa = dict()
        for checksum_addr, target_page, ctype, clength in packed_pages:
            root_node = ReFSChecksumChains.PageNode(checksum_addr, target_page, ctype, clength, self.mapping)
            if root_node.page_start_real in self._chains:
                self._chains[root_node.page_start_real].append(root_node)
            else:
                self._chains[root_node.page_start_real] = [root_node, ]

            if root_node.page_start in self._chains_pa:
                self._chains_pa[root_node.page_start].append(root_node)
            else:
                self._chains_pa[root_node.page_start] = [root_node, ]
            parent_nodes = self.__recursive_build(packed_pages, checksum_addr, self.mapping)
            root_node.parent_pages = parent_nodes


    def repair_checksums_ra(self, real_addr, mapped_image, testing = False): # type: (int, mmap.mmap, bool) -> None
        root_page_addr = self.mapping.get_page_start_ra(real_addr)
        root_nodes = self._chains.get(root_page_addr, None)
        if root_nodes is None:
            for ref in self.SUPB_self_ref:
                if root_nodes == None and ref.page_start_real == root_page_addr:
                    root_nodes = [ref,]
                    break

            for ref in self.CHKP_self_ref:
                if root_nodes == None and ref.page_start_real == root_page_addr:
                    root_nodes = [ref,]
                    break

        assert root_nodes != None, "No Checksum found for addr: 0x%s" % hex(real_addr)
        ReFSChecksumChains._repair_checksums(root_nodes, mapped_image, testing = testing)


    def repair_checksums_pa(self, packed_addr, mapped_image, testing = False): # type: (int, mmap.mmap) -> None
        root_page_addr = self.mapping.get_page_start(packed_addr)
        root_nodes = self._chains_pa.get(root_page_addr, None)
        if root_nodes is None:
            for ref in self.SUPB_self_ref:
                if root_nodes == None and ref.page_start == root_page_addr:
                    root_nodes = [ref,]
                    break

            for ref in self.CHKP_self_ref:
                if root_nodes == None and ref.page_start == root_page_addr:
                    root_nodes = [ref,]
                    break

        assert root_nodes != None, "No Checksum found for addr: 0x%s" % hex(packed_addr)
        ReFSChecksumChains._repair_checksums(root_nodes, mapped_image, use_packed = True, testing = testing)


    @staticmethod
    def _repair_checksum(node, mapped_image, use_packed = False, testing = False): # type: (ReFSChecksumChains.PageNode, mmap.mmap, bool) -> None
        if use_packed:
            target = node.page_start
            checksum_location = node.checksumed_at
        else:
            target = node.page_start_real
            checksum_location = node.checksumed_at_real

        target_size = node.page_size
        checksum_type = node.ctype
        checksum_length = node.length
        checksum_packed = None

        if checksum_type != ReFSChecksumChains.TYPE_CRC_32 and checksum_type != ReFSChecksumChains.TYPE_CRC_64:
            assert False, "Unknown checksum type"

        if checksum_type == ReFSChecksumChains.TYPE_CRC_32:
            if use_packed:
                assert node.self_ref_addr, "CRC32 can not be repaired without self descriptor data"
                self_ref_addr = node.self_ref_addr
            else:
                assert node.self_ref_addr_real, "CRC32 can not be repaired without self descriptor data"
                self_ref_addr = node.self_ref_addr_real
            assert node.self_ref_size, "CRC32 can not be repaired without self descriptor data"
            assert checksum_length == 4, "CRC32 must be length of 4"
            checksum = crc32castagnoli(target, target_size, self_ref_addr - target, node.self_ref_size, mapped_image)
            checksum_packed = pack("<L", checksum)


        if checksum_type == ReFSChecksumChains.TYPE_CRC_64:
            assert checksum_length == 8, "CRC64 must be length of 8"
            checksum = crc64ecma(target, target_size, mapped_image)
            checksum_packed = pack("<Q", checksum)


        # only for testing
        if testing:
            mapped_image.seek(checksum_location)
            checksum_data = mapped_image.read(checksum_length)
            assert len(checksum_packed) == len(checksum_data), "Checksum len mismatch: %d - %d" % (len(checksum_packed), len(checksum_data))
            for i in range(0, len(checksum_packed)):
                assert checksum_packed[i] == checksum_data[i], "Checksum calculation wrong"
            return

        mapped_image.seek(checksum_location)
        mapped_image.write(checksum_packed)


    @staticmethod
    def _repair_checksums(nodes, mapped_image, use_packed = False, testing = False): # type: (List[ReFSChecksumChains.PageNode], mmap.mmap, bool) -> None
        for node in nodes:
            ReFSChecksumChains._repair_checksum(node, mapped_image, use_packed, testing)
            ReFSChecksumChains._repair_checksums(node.parent_pages, mapped_image, use_packed, testing)


class ReFSDuplicateTables(object):
    def __init__(self, duplicate_tables_file): # type: (str) -> None
        self.__schema_table = list()
        self.__object_table = list()
        self.__container_table = list()

        self.__reparse_table = list()
        self.__volume_info_table = list()
        self.__log_info_table = list()
        self.__upcase_table = list()

        self.__init_duplicates(duplicate_tables_file)


    def get_schema(self):
        return tuple(self.__schema_table)

    def get_object(self):
        return tuple(self.__object_table)

    def get_container(self):
        return tuple(self.__container_table)

    def __init_duplicates(self, duplicate_tables_file): # type: (str) -> None
        with open(duplicate_tables_file, "r") as dtf:
            for line in dtf:
                orig_addr, dup_addr, name = parse("{} - {} - {}", line)
                orig_addr = int(orig_addr, 16)
                dup_addr = int(dup_addr, 16)
                if "SCHEMA" in name:
                    self.__schema_table.append(orig_addr)
                    self.__schema_table.append(dup_addr)
                if "OBJECT" in name:
                    self.__object_table.append(orig_addr)
                    self.__object_table.append(dup_addr)
                if "CONTAINER" in name:
                    self.__container_table.append(orig_addr)
                    self.__container_table.append(dup_addr)

                if "REPARSE" in name:
                    self.__reparse_table.append(orig_addr)
                    self.__reparse_table.append(dup_addr)
                if "VOLUME_INFO" in name:
                    self.__volume_info_table.append(orig_addr)
                    self.__volume_info_table.append(dup_addr)
                if "LOG_INFO" in name:
                    self.__log_info_table.append(orig_addr)
                    self.__log_info_table.append(dup_addr)
                if "UPCASE" in name:
                    self.__upcase_table.append(orig_addr)
                    self.__upcase_table.append(dup_addr)


    def correct_duplicate(self, page_addr, mapping, data): # type: (int, ReFSMapping, MMapInterface) -> Optional[int]
        duplicate_addr = self.get_duplicate(page_addr)
        if duplicate_addr is None:
            return None
        page_size = mapping.get_page_size(duplicate_addr)
        # at least address of page and table identifier are different in the duplicate: we will skip these values
        data.seek(page_addr)
        orig_data = data.read(page_size)
        orig_data_before_addr = orig_data[0:0x20]
        orig_data_after_addr = orig_data[0x50:]
        data.seek(duplicate_addr)
        data.write(orig_data_before_addr)
        data.seek(duplicate_addr + 0x50)
        data.write(orig_data_after_addr)
        return duplicate_addr


    def get_duplicate(self, page_addr, all_tables=True): # type: (int) -> Optional[int]
        if page_addr in self.__schema_table:
            index = self.__schema_table.index(page_addr) ^ 1
            return self.__schema_table[index]
        if page_addr in self.__object_table:
            index = self.__object_table.index(page_addr) ^ 1
            return self.__object_table[index]
        if page_addr in self.__container_table:
            index = self.__container_table.index(page_addr) ^ 1
            return self.__container_table[index]

        if all_tables and page_addr in self.__reparse_table:
            index = self.__reparse_table.index(page_addr) ^ 1
            return self.__reparse_table[index]
        if all_tables and page_addr in self.__volume_info_table:
            index = self.__volume_info_table.index(page_addr) ^ 1
            return self.__volume_info_table[index]
        if all_tables and page_addr in self.__log_info_table:
            index = self.__log_info_table.index(page_addr) ^ 1
            return self.__log_info_table[index]
        if all_tables and page_addr in self.__upcase_table:
            index = self.__upcase_table.index(page_addr) ^ 1
            return self.__upcase_table[index]
        return None



class ReFSExpander(object):
    def __init__(self, full_image_path, mapping = None): # type: (str, ReFSMapping) -> None
        if mapping is None:
            self.__mapping = _mapping
        else:
            self.__mapping = mapping

        assert self.__mapping is not None, "No mapping data available"
        self.__full_image_path = full_image_path


    def packed_to_full_image(self, payload): # type: (str) -> None
        translations = self.__mapping.get_addresses_to_expand()
        with open(self.__full_image_path, "r+b") as image_file:
            for poff, foff, size in translations:
                image_file.seek(foff)
                data = payload[poff:poff+size]
                image_file.write(data)


class ReFSPayload(object):

    class MMapInterface(array.array):
        def __new__(cls, data): # type: (str) -> None
            return super(ReFSPayload.MMapInterface, cls).__new__(cls, "B", data)

        def __init__(self, data):
            self.__index = 0

        def seek(self, index): # type: (int) -> None
            if index >= len(self):
                raise Exception
            self.__index = index

        def write(self, data): # type: (str) -> None
            data_len = len(data)
            self[self.__index:self.__index+data_len] = array.array("B", data)
            self.__index += data_len

        def read(self, length): # type: (int) -> str
            retval = self[self.__index:self.__index+length].tostring()
            self.__index += length
            return retval


    def __init__(self, old_payload, new_payload, checksums = None, mapping = None, dup_pages = None, correct_checksums = None, correct_duplicate_tables = None): # type: (str, str, ReFSChecksumChains, ReFSMapping, ReFSDuplicateTables, bool, bool) -> None
        assert len(old_payload) == len(new_payload), "Payload length has to be constant"
        self.packed_size = len(old_payload)
        if checksums:
            self.checksums = checksums
        else:
            self.checksums = _checksums
        if mapping:
            self.mapping = mapping
        else:
            self.mapping = _mapping

        if dup_pages:
            self.dup_pages = dup_pages
        else:
            self.dup_pages = _duplicate_tables

        if correct_checksums is None:
            self.correct_checksums = _correct_checksums
        else:
            self.correct_checksums = correct_checksums

        if correct_duplicate_tables is None:
            self.correct_duplicate_tables = _correct_duplicate_tables
        else:
            self.correct_duplicate_tables = correct_duplicate_tables

        assert self.checksums is not None, "No checksum data available"
        assert self.mapping is not None, "No mapping data available"
        self.__old_payload = old_payload
        self.__new_payload = new_payload

    def get_corrected_payload(self): # type: () -> str
        data = self.MMapInterface(self.__new_payload)

        modified_pages = self.__get_modified_pages()

        if self.correct_duplicate_tables:
            self.__duplicate_pages(modified_pages, data)
            self.__add_duplicated_pages(modified_pages)

        if self.correct_checksums:
            for addr in modified_pages:
                self.checksums.repair_checksums_pa(addr, data)

        return data.tostring()


    def __get_modified_pages(self): # type: () -> Set[int]
        mod_pages = set()
        for i in range(0, self.packed_size):
            if self.__old_payload[i] != self.__new_payload[i]:
                # maybe we can optimize here and only call get_page_start for diff begin and diff end
                mod_pages.add(self.mapping.get_page_start(i))
        return mod_pages


    def __add_duplicated_pages(self, page_addr): # type: (Set[int]) -> None
        mod_pages = list()
        for paddr in page_addr:
            duplicate_addr = self.dup_pages.get_duplicate(paddr)
            if duplicate_addr is not None:
                mod_pages.append(duplicate_addr)
        page_addr.update(mod_pages)


    def __duplicate_pages(self, page_addr, data): # type: (Set[int], MMapInterface): -> None
        duplicated = set()
        for paddr in page_addr:
            if paddr in duplicated:
                continue
            daddr = self.dup_pages.correct_duplicate(paddr, self.mapping, data)
            if daddr is not None:
                duplicated.add(daddr)
